@@ -49,7 +49,7 @@
 
 ## modifies settings:
 
-rc.settings <- function(ops, ns, args, func, ipck, S3, data, help, argdb)
+rc.settings <- function(ops, ns, args, func, ipck, S3, data, help, argdb, files)
 {
     checkAndChange <- function(what, value)
     {
@@ -58,15 +58,16 @@ rc.settings <- function(ops, ns, args, func, ipck, S3, data, help, argdb)
             !is.na(value))
             .CompletionEnv$settings[[what]] <- value
     }
-    if (!missing(ops))   checkAndChange( "ops",  ops)
-    if (!missing(ns))    checkAndChange(  "ns",   ns)
-    if (!missing(args))  checkAndChange("args", args)
-    if (!missing(func))  checkAndChange("func", args)
-    if (!missing(ipck))  checkAndChange("ipck", ipck)
-    if (!missing(S3))    checkAndChange("S3", S3)
-    if (!missing(data))  checkAndChange("data", S3)
-    if (!missing(help))  checkAndChange("help", S3)
-    if (!missing(argdb)) checkAndChange("argdb", S3)
+    if (!missing(ops))   checkAndChange(  "ops",   ops)
+    if (!missing(ns))    checkAndChange(   "ns",    ns)
+    if (!missing(args))  checkAndChange( "args",  args)
+    if (!missing(func))  checkAndChange( "func",  args)
+    if (!missing(ipck))  checkAndChange( "ipck",  ipck)
+    if (!missing(S3))    checkAndChange(   "S3",    S3)
+    if (!missing(data))  checkAndChange( "data",  data)
+    if (!missing(help))  checkAndChange( "help",  help)
+    if (!missing(argdb)) checkAndChange("argdb", argdb)
+    if (!missing(files)) checkAndChange("files", files)
     invisible()
 }
 
@@ -129,7 +130,7 @@ rc.status <- function()
 .assignEnd           <- function(end)   assign("end",        end,   envir = .CompletionEnv)
 .setFileComp         <- function(state) assign("fileName",   state, envir = .CompletionEnv)
 
-.retrieveCompletions <- function()         get("comps",             envir = .CompletionEnv)
+.retrieveCompletions <- function()  unique(get("comps",             envir = .CompletionEnv))
 .getFileComp         <- function()         get("fileName",          envir = .CompletionEnv)
 
 
@@ -142,11 +143,28 @@ rc.status <- function()
 
 
 .guessTokenFromLine <-
-    function()
+    function(linebuffer = .CompletionEnv[["linebuffer"]],
+             end = .CompletionEnv[["end"]])
 {
-    linebuffer <- .CompletionEnv[["linebuffer"]]
-    end <- .CompletionEnv[["end"]]
-    start <- suppressWarnings(gregexpr("[^\\.\\w:?$@[\\]]+", substr(linebuffer, 1, end), perl = TRUE))[[1]]
+    ## special rules apply when we are inside quotes (see fileCompletionPreferred() below)
+    insideQuotes <- {
+        lbss <- head(unlist(strsplit(linebuffer, "")), .CompletionEnv[["end"]])
+        ((sum(lbss == "'") %% 2 == 1) ||
+         (sum(lbss == '"') %% 2 == 1))
+    }
+    start <- 
+        if (insideQuotes)
+
+            ## FIXME (easy): should just set 'start' to the location
+            ## of the last quote, but that would need more minutes of
+            ## thinking than I have right now.
+
+            suppressWarnings(gregexpr("[^\\.\\w:?$@[\\]\\\\/~ ]+", substr(linebuffer, 1, end), perl = TRUE))[[1]]
+        else
+            suppressWarnings(gregexpr("[^\\.\\w:?$@[\\]]+", substr(linebuffer, 1, end), perl = TRUE))[[1]]
+
+    
+            
     ##                                    ^^^^^^^^^^^^^^
     ##                             things that should not cause breaks
     start <- ## 0-indexed
@@ -503,8 +521,6 @@ inFunction <-
         prefix <- substr(line, 1, index - 1)
         suffix <- substr(line, index + 1, cursor + 1)
 
-
-
         ## note in passing whether we are the first argument (no '='
         ## and no ',' in suffix)
 
@@ -518,7 +534,8 @@ inFunction <-
                                      tail(gregexpr("=", suffix, fixed = TRUE)[[1]], 1),
                                      1e6), fixed = TRUE)) == 0))
         {
-            ## we are on the wrong side of a = to be an argument
+            ## we are on the wrong side of a = to be an argument, so
+            ## we don't care even if we are inside a function
             return(character(0))
         }
         else ## guess function name
@@ -646,19 +663,46 @@ functionArgs <-
 ## number of quotes between the cursor and the beginning of the line
 ## is an odd number.
 
+isInsideQuotes <- 
 fileCompletionPreferred <- function()
 {
     ((st <- .CompletionEnv[["start"]]) > 0 && {
         
+        ## yes if the number of quote signs to the left is odd
         linebuffer <- .CompletionEnv[["linebuffer"]]
         lbss <- head(unlist(strsplit(linebuffer, "")), .CompletionEnv[["end"]])
         ((sum(lbss == "'") %% 2 == 1) ||
          (sum(lbss == '"') %% 2 == 1))
-
+        
     })
 }
 
 
+## File name completion, used if settings$files == TRUE.  Front ends
+## that can do filename completion themselves should probably not use
+## this as they will do a better job.
+
+
+fileCompletions <- function(token)
+{
+    ## uses Sys.glob (conveniently introduced in 2.5.0)
+    ## assume token starts just after the begin quote
+
+    ## Sys.glob doesn't work without expansion.  Is that intended?
+    token.expanded <- path.expand(token)
+    comps <- Sys.glob(sprintf("%s*", token.expanded), dirmark = TRUE)
+
+    ## for things that only extend beyond the cursor, need to
+    ## 'unexpand' path
+    if (token.expanded != token)
+        comps <- sub(path.expand("~"), "~", comps, fixed = TRUE)
+    comps
+}
+
+
+
+
+
 ## .completeToken() is the primary interface, and does the actual
 ## completion when called from C code.
 
@@ -666,23 +710,53 @@ fileCompletionPreferred <- function()
 .completeToken <- function()
 {
     text <- .CompletionEnv[["token"]]
-    if (fileCompletionPreferred())
+    if (isInsideQuotes())
     {
 
-        ## If we're in here, that means we think standard filename
-        ## completion is more appropriate (by design, this is supposed
-        ## to happen when we're inside quotes, which is not exactly
-        ## equivalent, but there's not much else we can do anyway in
-        ## that case).  We make no attempt to do filename completion
-        ## because other people have written better code to do that,
-        ## and so we set our completion list to be empty.  Third party
-        ## software using this code can interrogate
-        ## rc.status("fileName") to determine if this is the situation
-        ## and act accordingly.  It's probably even OK to fill
-        ## .CompletionEnv$comps with something suitable.
+        ## If we're in here, that means we think the cursor is inside
+        ## quotes.  In most cases, this means that standard filename
+        ## completion is more appropriate, but probably not if we're
+        ## trying to access things of the form x["foo... or x$"foo...
+        ## The following tries to figure this out, but it won't work
+        ## in all cases (e.g. x[, "foo<TAB>"])
 
-        .CompletionEnv[["comps"]] <- character(0)
-        .setFileComp(TRUE)
+        ## We assume that whoever determines our token boundaries
+        ## considers quote signs as a breaking symbol.
+
+        st <- .CompletionEnv[["start"]]
+        probablyNotFilename <- 
+            (st > 2 && 
+             (substr(.CompletionEnv[["linebuffer"]], st-1, st-1) %in% c("[", ":", "$")))
+
+
+        ## If the 'files' setting is FALSE, we will make no attempt to
+        ## do filename completion (this is likely to happen with
+        ## front-ends that are capable of doing their own file name
+        ## completion; such front-ends can fall back to their native
+        ## file completion when rc.status("fileName") is TRUE.
+
+        if (.CompletionEnv$settings[["files"]])
+        {
+
+            ## we bail out if probablyNotFilename == TRUE.  If we
+            ## wanted to be really fancy, we could try to detect where
+            ## we are and act accordingly, e.g. if it's
+            ## foo[["bar<TAB>, pretend we are completing foo$bar, etc.
+            ## This is not that hard, we just need to use
+            ## .guessTokenFromLine(linebuffer, end = st - 1) to get the
+            ## 'foo[[' part.
+
+            .CompletionEnv[["comps"]] <-
+                if (probablyNotFilename) character(0)
+                else fileCompletions(text)
+            .setFileComp(FALSE)
+        }
+        else
+        {
+            .CompletionEnv[["comps"]] <- character(0)
+            .setFileComp(TRUE)
+        }
+        
     }
     else
     {
@@ -702,12 +776,12 @@ fileCompletionPreferred <- function()
 
         fargComps <- functionArgs(guessedFunction, text)
 
-        if (getIsFirstArg() &&
+        if (getIsFirstArg() && length(guessedFunction) > 0 &&
             guessedFunction %in%
             c("library", "require", "data"))
         {
-            ## don't try anything else
             .CompletionEnv[["comps"]] <- fargComps
+            ## don't try anything else
             return()
         }
 
